@@ -12,7 +12,8 @@ test_that("Monthly, in_type=='params', distr='gaussian'",{
   # Run Gauss Reconc
   res.gauss = reconc_gaussian(S, base_forecasts_in[[1]], diag(base_forecasts_in[[2]]^2))
   # Test
-  m = mean(rowMeans(res.buis$reconciled_samples) - as.numeric(rbind(res.gauss$upper_reconciled_mean, res.gauss$bottom_reconciled_mean)))
+  b_mask = rowSums(S) == 1
+  m = mean(rowMeans(res.buis$reconciled_samples)[b_mask] - as.numeric(res.gauss$bottom_reconciled_mean))
   expect_equal(abs(m) < 8e-3, TRUE)
 })
 
@@ -30,7 +31,8 @@ test_that("Weekly, in_type=='params', distr='gaussian'",{
   # Run Gauss Reconc
   res.gauss = reconc_gaussian(S, base_forecasts_in[[1]], diag(base_forecasts_in[[2]]^2))
   # Test
-  m = mean(rowMeans(res.buis$reconciled_samples) - as.numeric(rbind(res.gauss$upper_reconciled_mean, res.gauss$bottom_reconciled_mean)))
+  b_mask = rowSums(S) == 1
+  m = mean(rowMeans(res.buis$reconciled_samples)[b_mask] - as.numeric(res.gauss$bottom_reconciled_mean))
   expect_equal(abs(m) < 2e-2, TRUE)
 })
 
@@ -47,6 +49,22 @@ test_that("Monthly, in_type=='params', distr='poisson'",{
   expect_no_error(res.buis)
 })
 
+
+test_that("Monthly, in_type=='params', distr='nbinom'",{
+  S = read.csv(file = "dataForTests/Monthly-NegBin_S.csv", header = FALSE)
+  S = as.matrix(S)
+  base_forecasts_in = read.csv(file = "dataForTests/Monthly-NegBin_basef.csv", header = FALSE)
+  base_forecasts = list()
+  for (i in 1:nrow(base_forecasts_in)) {
+    base_forecasts[[i]] = list(as.numeric(base_forecasts_in[i,]))[[1]]
+  }
+  res.buis = reconc_BUIS(S, base_forecasts,
+                         in_type = "params", distr = "nbinom", num_samples = 10000, seed=42)
+  expect_no_error(res.buis)
+})
+
+
+
 test_that("Monthly, in_type=='samples', distr='continuous'",{
   # Run IS Reconc from samples
   S = read.csv(file = "dataForTests/Monthly-Gaussian_S.csv", header = FALSE)
